@@ -175,7 +175,7 @@ class AntecedentEstimator(object):
         
         # Input:
         # Xin: N x 1 input domain (column vector)
-        # MFin: N x 1correspoding membership values 
+        # MFin: N x 1 corresponding membership values 
         # nc: number of alpha cut values to consider (default=101)
         # norm: optional normalization flag (0: do not normalize, 1 : normalize, 
         # default=1)
@@ -207,6 +207,17 @@ class AntecedentEstimator(object):
                 mf[i]=acut[i]
                 mf[i+nc]=acut[i]
                 
+        # # Determine the elements in the alpha cuts    
+        # for i in range(0,nc):
+        #     tmp1 = mfin>acut[i]
+        #     if len(tmp1)==0:
+        #         tmp=xin[tmp1]
+        #         np.sort(tmp)
+        #         x[i]=tmp[0]
+        #         x[i+nc]=tmp[-1]
+        #         mf[i]=acut[i]
+        #         mf[i+nc]=acut[i]
+                  
         #Delete NaNs
         idx=np.isnan(x)
         x=x[idx==False]
@@ -270,18 +281,31 @@ class AntecedentEstimator(object):
             
             # Fit parameters to the data using least squares
 #            print('mu1',mu1,'sig1',sig1,'mu2',mu2,'sig2',sig2)
-            param, _ = curve_fit(self._gauss2mf, x, mf, p0 = [mu1, sig1, mu2, sig2], bounds=((-np.inf, 0), (np.inf, np.inf)), maxfev=1000)
+            param, _ = curve_fit(self._gauss2mf, x, mf, p0 = [mu1, sig1, mu2, sig2], maxfev=1000, bounds=((-np.inf, 0,-np.inf, 0), (np.inf, np.inf,np.inf, np.inf)))
             
         elif mf_shape == 'sigmf':
             # Determine initial parameters
             if np.argmax(mf)-np.argmin(mf) > 0:         # if sloping to the right
-                c = x[mf>=0.5][0]
-                s = 1
+                if len(x[mf>=0.5])>0:
+                    c = x[mf>=0.5][0]
+                    s = 1
+                elif len(x[mf>=0.5])==0:                   # if there are no datapoints with membership larger than 0
+                    c = x[0]
+                    s = 1                    
             elif np.argmax(mf)-np.argmin(mf) < 0:       # if sloping to the left
-                c = x[mf<=0.5][0]
-                s = -1
-            # Fit parameters of the function to the data using non-linear least squares           
-            param, _ = curve_fit(self._sigmf, x, mf, p0 = [c, s])
+                if len(x[mf<=0.5])>0:
+                    c = x[mf<=0.5][0]
+                    s = 1
+                elif len(x[mf<=0.5])==0:                   # if there are no datapoints with membership smaller than 0
+                    c = x[-1]
+                    s = 1             
+            # Fit parameters of the function to the data using non-linear least squares  
+            try:
+                param, _ = curve_fit(self._sigmf, x, mf, p0 = [c, s], maxfev=1000)
+            except RuntimeError:
+                print('pyFUME attempted to fit sigmoidal shaped membership functions, but was unable to find fitting parameters. pyFUME will now terminate. Please consider using a different shape for the membership functions.')
+                import sys
+                sys.exit()
         
         return mf_shape, param
         
