@@ -31,7 +31,7 @@ class pyFUME(object):
             An object containing the fuzzy model, information about its setting (such as its antecedent and consequent parameters) and the different splits of the data.
     """
     def __init__(self, datapath=None, dataframe=None, nr_clus=2, process_categorical=False, method='Takagi-Sugeno', variable_names=None, merge_threshold=1., **kwargs):
-
+      
         if datapath is None and dataframe is None:
             raise Exception("ERROR: a dataset was not specified. Please either use the datapath or dataframe arguments.")
 
@@ -150,8 +150,10 @@ class pyFUME(object):
         if self.FIS.minmax_norm_flag==True:
             if np.amin(data)<0 or np.amax(data) >1:
                 print('WARNING: The given value(s) are not between 0 and 1, the denormalization is performed by extrapolating.')
+            
             norm_val=self.FIS.normalization_values
             variable_names, min_values, max_values = zip(*norm_val)
+            # print(min_values, max_values)
             denormalized_data = (data * (np.array(max_values) - np.array(min_values))) + np.array(min_values)
         elif self.FIS.minmax_norm_flag==False:
             raise Exception('The model was not trained on normalized data, normalization is aborted.')    
@@ -188,8 +190,7 @@ class pyFUME(object):
             The fuzzy model (as an executable object).
         """          
         if self.FIS.model is None:
-            print ("ERROR: model was not created correctly, aborting.")
-            exit(-1)
+            raise Exception("ERROR: model was not created correctly, aborting.")
         else:
             return self.FIS.model
         
@@ -258,6 +259,25 @@ class pyFUME(object):
         MAPE = test.calculate_MAPE()
         return MAPE
     
+    def _get_accuracy(self):
+        # Calculate the accuraccy of the model using the test data set
+        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, golden_standard=self.FIS.y_test, variable_names=self.FIS.variable_names)
+        accuracy = test.calculate_accuracy()
+        return accuracy
+    
+    def calculate_AUC(self, number_of_slices=100, show_plot = False):
+        # Calculate the area under the ROC curve of the model using the test data set
+        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, golden_standard=self.FIS.y_test, variable_names=self.FIS.variable_names)
+        AUC = test.calculate_AUC(number_of_slices, show_plot)
+        return AUC
+
+    
+    def get_confusion_matrix(self):
+        # Calculate the confusion matrix of the model using the test data set
+        test = SugenoFISTester(model=self.FIS.model, test_data=self.FIS.x_test, golden_standard=self.FIS.y_test, variable_names=self.FIS.variable_names)
+        con_mat = test.generate_confusion_matrix()
+        return con_mat
+    
     def get_data(self, data_set='test'):
         """
         Returns the test or training data set.
@@ -294,7 +314,7 @@ class pyFUME(object):
     ### PLOT FACILITIES ###
     #######################
 
-    def plot_mf(self, variable_name, output_file='', highlight_element=None, highlight_mf=None, ax = None):
+    def plot_mf(self, variable_name, output_file='', highlight_element=None, highlight_mf=None, ax = None, xscale = 'linear'):
         """
         Uses Simpful's plotting facilities to plot the membership functions of
         the pyFUME model.
@@ -307,7 +327,7 @@ class pyFUME(object):
             ax: The motplotlib ax where the variable will be plotted.
 
         """ 
-        self.get_model().plot_variable(var_name = variable_name, outputfile = output_file, TGT = highlight_element, highlight = highlight_mf, ax = ax)
+        self.get_model().plot_variable(var_name = variable_name, outputfile = output_file, TGT = highlight_element, highlight = highlight_mf, ax = ax, xscale = xscale)
                 
     def plot_all_mfs(self, output_file='', figures_per_row=4):
         """
@@ -409,19 +429,23 @@ class pyFUME(object):
         """        
     
         (_, min_value, max_value) = normalization_values
-        tmp = []
-        for x in data[-1]:
-            denormalized_data = (x * (np.array(max_value) - np.array(min_value))) + np.array(min_value)
-            tmp.append(denormalized_data)
-            
-        denormalized_set = tuple([data[0], tmp])
+
+        x=data[-1]
+
+        denormalized_mu = (x[0] * (np.array(max_value) - np.array(min_value))) + np.array(min_value)
+        denormalized_sigma = (x[1] * (np.array(max_value) - np.array(min_value))) 
+
+        denormalized_set = tuple([data[0], [denormalized_mu, denormalized_sigma]])
 
         return denormalized_set
     
-    def plot_denormalized_mf(self, variable_name, output_file='', highlight_element=None, highlight_mf=None, ax = None):
+    def plot_denormalized_mf(self, variable_name, output_file='', highlight_element=None, highlight_mf=None, ax = None, xscale = 'linear'):
         normalization_values = self.FIS.normalization_values 
         antecedent_sets= self.FIS.antecedent_parameters
         
+        if normalization_values == None:
+            raise Exception('ERROR: The input data for the pyFUME model was not normalized during training. Denormaliztaion is therefore not possible.')
+        
         # Check if variables were removed (during feature selection)
         to_keep=[]
         for i in range(0,len(normalization_values)):
@@ -433,13 +457,18 @@ class pyFUME(object):
         
         # Denormalize the antecedent set parameters
         denormed_antecedent_sets = []
+        
+        x=0
+        cnt=0
         for i in range(0, len(antecedent_sets)):
-            if i in range(0,len(self.FIS.selected_variable_names)):
-                norm_vals=normalization_values[i]
-            else:
-                norm_vals=normalization_values[i-len(self.FIS.selected_variable_names)]
-            denormed_set=self._denormalize_antecedent_set(antecedent_sets[i], norm_vals)
-            denormed_antecedent_sets.append(denormed_set)
+             norm_vals = normalization_values[x]
+             denormed_set=self._denormalize_antecedent_set(antecedent_sets[i], norm_vals)
+             denormed_antecedent_sets.append(denormed_set)
+             cnt+=1
+            
+             if cnt == self.nr_clus:
+                x+=1
+                cnt = 0
         
         UoD = []
         _, mi, ma = zip(*self.FIS.normalization_values)
@@ -459,11 +488,8 @@ class pyFUME(object):
         dummymodel = simpbuilder.simpfulmodel
         
         # Plot the requested variable using Simpful
-        dummymodel.plot_variable(var_name = variable_name, outputfile = output_file, TGT = highlight_element, highlight = highlight_mf, ax = ax)
-
-        
-
-
+        dummymodel.plot_variable(var_name = variable_name, outputfile = output_file, TGT = highlight_element, highlight = highlight_mf, ax = ax, xscale = xscale)
+      
 if __name__=='__main__':
     from numpy.random import seed
     seed(4)
